// EXTERNAL LINKER DEPENDENCIES (codeblocks+gcc)
// liballegro-5.0.7-monolith-md-debug.a
// liballegro-5.0.7-monolith-mt.a

#include <iostream>
#include <math.h>

#include <allegro5\allegro.h>
#include <allegro5\allegro_font.h>
#include <allegro5\allegro_ttf.h>
#include <allegro5\allegro_primitives.h>
#include <allegro5\allegro_image.h>
#include <allegro5\allegro_audio.h>
#include <allegro5\allegro_acodec.h>

using namespace std;

int get_config_int(const ALLEGRO_CONFIG *config, const char *section, const char *key, int default_value)
{
	const char * buffer = al_get_config_value(config, section, key);
	if(buffer)
        return atoi(buffer);
	else
        return default_value;
}

float get_config_float(const ALLEGRO_CONFIG *config, const char *section, const char *key, float default_value)
{
	const char * buffer = al_get_config_value(config, section, key);
	if(buffer)
        return atof(buffer);
	else
        return default_value;
}

float get_sample_amplitude (ALLEGRO_SAMPLE *sample, int position)
{
    // Return a float value from 0 to 1.
    short *spl_data = (short*)al_get_sample_data(sample);
    return float(sqrt(pow(spl_data[position], 2))/32768.0f);
}

int main ()
{
	bool exit = false;
	bool rendering = true;

	cout << "Starting Allegro5...";

	// ALLEGRO INIT
	if (!al_init())
	{
		cout << "\nUnable to start Allegro5!\n\n";
		system("pause");
		return -1;
	}
	else
		cout << "OK!\n\n";

	// LOAD CONFIG FILE
	cout << "Loading file config.cfg...";
	ALLEGRO_CONFIG *config_file = al_load_config_file("data\\config.cfg");
	cout << "OK!\n\n";

	// VIDEO INIT
	bool fullscreen = get_config_int(config_file, "DISPLAY", "fullscreen", 0);
	bool hide_mouse = get_config_int(config_file, "DISPLAY", "hide_mouse", 1);
	int FPS = get_config_int(config_file, "DISPLAY", "fps", 60);
    int frames = 0, game_time = 0, game_FPS = 0;

	cout << "Detecting screen info...\n";
	ALLEGRO_MONITOR_INFO user_monitor_info;
	al_get_monitor_info(0, &user_monitor_info);
	int user_display_width = user_monitor_info.x2 - user_monitor_info.x1;
	int user_display_height = user_monitor_info.y2 - user_monitor_info.y1;

	cout << "Your screen is: " << user_display_width << " x " << user_display_height << endl;

	int display_width = 0;
	int display_height = 0;

	if (get_config_int(config_file, "DISPLAY", "display_width", 800) == 0)
		display_width = user_display_width;
	else
		display_width = get_config_int(config_file, "DISPLAY", "display_width", 800);

	if (get_config_int(config_file, "DISPLAY", "display_height", 600) == 0)
		display_height = user_display_height;
	else
		display_height = get_config_int(config_file, "DISPLAY", "display_height", 600);

	if (fullscreen)
		al_set_new_display_flags(ALLEGRO_FULLSCREEN);

	cout << display_width << " x " << display_height << " resolution setted.\n\n";

    // Setting anti-aliasing for primitives
	//al_set_new_display_option(ALLEGRO_SAMPLE_BUFFERS, 1, ALLEGRO_SUGGEST);
	//al_set_new_display_option(ALLEGRO_SAMPLES, 2, ALLEGRO_SUGGEST);

	cout << "Creating display...";
	ALLEGRO_DISPLAY *display = al_create_display(display_width, display_height);
	ALLEGRO_TIMER *fps_timer = al_create_timer(1.0 / FPS);

	if (!display)
	{
		cout << "Allegro5: failed to create display!\n\n";
		system("pause");
		return -1;
	}
	else
		cout << "OK!\n\n";

    if (hide_mouse)
		al_hide_mouse_cursor(display);

	// INIT ADDONS
	al_install_keyboard();
	al_install_mouse();
	al_init_font_addon();
	al_init_ttf_addon();
	al_init_primitives_addon();
    al_init_image_addon();
	al_install_audio();
    al_init_acodec_addon();

	// FONT
	const char *font_file = al_get_config_value(config_file, "DISPLAY", "font");
	cout << "Loading font...\n";
	cout << font_file << endl << endl;
	int default_text_size = get_config_int(config_file, "DISPLAY", "default_text_size", 40);
	int fps_text_size = get_config_int(config_file, "DISPLAY", "fps_text_size", 60);
	int key_text_size = get_config_int(config_file, "DISPLAY", "key_text_size", 30);
	ALLEGRO_FONT *default_text = al_load_font(font_file, default_text_size, 0);
	ALLEGRO_FONT *fps_text = al_load_font(font_file, fps_text_size, 0);
	ALLEGRO_FONT *key_text = al_load_font(font_file, key_text_size, 0);

	// BITMAPS
	cout << "Loading bitmaps...\n";
	cout << "data\\cursor.png" << endl << endl;
	ALLEGRO_BITMAP *original_cursor = al_load_bitmap("data\\cursor.png");
    int original_cursor_width = al_get_bitmap_width(original_cursor);
	int original_cursor_height = al_get_bitmap_height(original_cursor);
	int cursor_rad = (get_config_int(config_file, "DISPLAY", "cursor_size", 40))/2;
	ALLEGRO_BITMAP *cursor = al_create_bitmap(cursor_rad * 2, cursor_rad * 2);
	al_set_target_bitmap(cursor);
	al_draw_scaled_bitmap(original_cursor, 0, 0, original_cursor_width, original_cursor_height, 0, 0, cursor_rad * 2, cursor_rad * 2, 0);
	al_set_target_bitmap(al_get_backbuffer(display));

	// AUDIO
	const char *song_path = al_get_config_value(config_file, "AUDIO", "song_path");
	float music_volume = get_config_float(config_file, "AUDIO", "music_volume", 0.6);
	float samples_volume = get_config_float(config_file, "AUDIO", "sounds_volume", 0.8);

    al_reserve_samples(20); // Mixer channels
	cout << "Loading sounds files...\n";
	cout << song_path << endl;
	ALLEGRO_SAMPLE *song = al_load_sample(song_path);
	ALLEGRO_SAMPLE_INSTANCE *song_instance = al_create_sample_instance(song);
	al_set_sample_instance_gain(song_instance, music_volume);
	al_set_sample_instance_playmode(song_instance, ALLEGRO_PLAYMODE_LOOP);
	al_attach_sample_instance_to_mixer(song_instance, al_get_default_mixer());
	al_play_sample_instance(song_instance);

	ALLEGRO_SAMPLE *sample[8];
	ALLEGRO_SAMPLE_INSTANCE *sample_instance[8];
	char path[30] = "data\\samples\\sample (0).wav";

	for (int i = 0; i < 8; i++)
	{
		path[21] = '0'+(i + 1);
		cout << path << endl;
		sample[i] = al_load_sample(path);
		sample_instance[i] = al_create_sample_instance(sample[i]);
		al_set_sample_instance_gain(sample_instance[i], samples_volume);
		al_set_sample_instance_playmode(sample_instance[i], ALLEGRO_PLAYMODE_ONCE);
		al_attach_sample_instance_to_mixer(sample_instance[i], al_get_default_mixer());
	}

	// SERVICES
	ALLEGRO_EVENT_QUEUE *event_queue = al_create_event_queue();
	al_register_event_source(event_queue, al_get_display_event_source(display));
	al_register_event_source(event_queue, al_get_timer_event_source(fps_timer));
	al_register_event_source(event_queue, al_get_mouse_event_source());
	al_register_event_source(event_queue, al_get_keyboard_event_source());

	al_start_timer(fps_timer);
	game_time = al_current_time();

	// VARIABLES
	ALLEGRO_COLOR background_color = al_map_rgb(0, 0, 0);

    const int sample_key[8] = {ALLEGRO_KEY_Z, ALLEGRO_KEY_X, ALLEGRO_KEY_C, ALLEGRO_KEY_V, ALLEGRO_KEY_B, ALLEGRO_KEY_N, ALLEGRO_KEY_M, ALLEGRO_KEY_COMMA};
    char sample_displayed_key[8] = {'z', 'x', 'c', 'v', 'b', 'n', 'm', ','};
	bool sample_key_pressed[8] = {0, 0, 0, 0, 0, 0, 0, 0};
	bool trigger[8] = {0, 0, 0, 0, 0, 0, 0, 0};
	bool meter_animation[8] = {0, 0, 0, 0, 0, 0, 0, 0};
    bool song_play = true;

    int pos_x = (display_width/2) - cursor_rad;
	int pos_y = (display_height/2) - cursor_rad;
    float dinamic_pitch = 1.0;

    ALLEGRO_COLOR meter_colors[8] = {al_map_rgb(196, 12, 10), al_map_rgb(249, 123, 12), al_map_rgb(247, 234, 0), al_map_rgb(6, 222, 17), al_map_rgb(4, 218, 164), al_map_rgb(24, 36, 254), al_map_rgb(88, 2, 182), al_map_rgb(146, 1, 106)};
	const int meter_orizzontal_offset = 20;
	const int meter_down_offset = 100;
	const int meter_top_offset = 100;
    const int meter_default_height = 20;
	int meter_width = (display_width - (meter_orizzontal_offset*2)) / 8;
	int meter_max_height = display_height - (meter_top_offset + meter_down_offset);
	int meter_thickness = 10;

    cout << endl << "Game loop!\n\n";

	// MAIN LOOP
	while (!exit)
	{
		ALLEGRO_EVENT current_event;
		al_wait_for_event(event_queue, &current_event);

		if(current_event.type == ALLEGRO_EVENT_DISPLAY_CLOSE)
			exit = true;

		if(current_event.type == ALLEGRO_EVENT_KEY_UP && current_event.keyboard.keycode == ALLEGRO_KEY_ESCAPE)
			exit = true;

		if (current_event.type == ALLEGRO_EVENT_KEY_DOWN)
		{
			for (int i = 0; i < 8; i++)
			{
				if (current_event.keyboard.keycode == sample_key[i])
				{
					sample_key_pressed[i] = true;
				}
			}

			if (current_event.keyboard.keycode == ALLEGRO_KEY_SPACE)
				song_play = !song_play;
		}

		if (current_event.type == ALLEGRO_EVENT_KEY_UP)
		{
			for (int i = 0; i < 8; i++)
			{
				if (current_event.keyboard.keycode == sample_key[i])
					sample_key_pressed[i] = false;
			}
		}

		if (current_event.type == ALLEGRO_EVENT_MOUSE_AXES)
		{
			pos_x = current_event.mouse.x;
			pos_y = current_event.mouse.y;

			dinamic_pitch = ((float)display_height-10)/(float)pos_y - 1;
		}

		if (current_event.type == ALLEGRO_EVENT_TIMER)
		{
			frames++;

			if(al_current_time() - game_time >= 1)
			{
				game_time = al_current_time();
				game_FPS = frames;
				frames = 0;
			}

			rendering = true;
		}

		for (int i = 0; i < 8; i++)
		{
			if (sample_key_pressed[i])
			{
				if (!trigger[i])
				{
					trigger[i] = true;

					if (!al_get_sample_instance_playing(sample_instance[i]))
					{
						al_set_sample_instance_speed(sample_instance[i], (float)dinamic_pitch);
						al_play_sample_instance(sample_instance[i]);
						cout << "Play sample ("<< i + 1 << ").wav at " << (int)(al_get_sample_instance_speed(sample_instance[i])*100) << "% of speed"<< endl;
					}
				}
			}
			else
			{
				trigger[i] = false;

				if (al_get_sample_instance_playing(sample_instance[i]))
					al_stop_sample_instance(sample_instance[i]);
			}

			if (al_get_sample_instance_playing(sample_instance[i]))
				meter_animation[i] = true;
			else
				meter_animation[i] = false;
		}

		if (rendering && al_is_event_queue_empty(event_queue))
		{
            // Draw meters
			for (int i = 0; i < 8; i++)
			{
				if (meter_animation[i])
				{
				    // Meter animation
					al_draw_rectangle(meter_orizzontal_offset + (meter_width * i), display_height - meter_down_offset, meter_orizzontal_offset + meter_width + (meter_width * i), (display_height - meter_down_offset - (meter_max_height*get_sample_amplitude(sample[i], al_get_sample_instance_position(sample_instance[i])))) - meter_default_height, meter_colors[i], meter_thickness);

					al_draw_textf(key_text, meter_colors[i], (meter_width/2) + (meter_orizzontal_offset + (meter_width * i)), display_height - meter_down_offset, ALLEGRO_ALIGN_CENTRE, "%c", sample_displayed_key[i]);
				}
				else
				{
					al_draw_rectangle(meter_orizzontal_offset + (meter_width * i), display_height - meter_down_offset, meter_orizzontal_offset + meter_width + (meter_width * i), (display_height - meter_down_offset) - meter_default_height, meter_colors[i], meter_thickness);

					al_draw_textf(key_text, al_map_rgb(255, 255, 255), (meter_width/2) + (meter_orizzontal_offset + (meter_width * i)), display_height - meter_down_offset, ALLEGRO_ALIGN_CENTRE, "%c", sample_displayed_key[i]);
				}
			}

            if (!al_get_sample_instance_playing(song_instance) & song_play)
			{
				cout << "PLAY song.wav...\n";
				al_play_sample_instance(song_instance);

			}
			else if (al_get_sample_instance_playing(song_instance) & !song_play)
			{
				cout << "STOP song.wav...\n";
				al_stop_sample_instance(song_instance);
			}


            if (song_play)
                al_draw_text(key_text, al_map_rgb(100, 100, 100), display_width, display_height - key_text_size - 10, ALLEGRO_ALIGN_RIGHT, "space to STOP song");
            else
                al_draw_text(key_text, al_map_rgb(100, 100, 100), display_width, display_height - key_text_size - 10, ALLEGRO_ALIGN_RIGHT, "space to PLAY song");

			al_draw_textf(default_text, al_map_rgb(255, 255, 255), 0, 0, ALLEGRO_ALIGN_LEFT, "SamplePlayer ep");
            al_draw_textf(fps_text, al_map_rgb(255, 255, 255), display_width, 0, ALLEGRO_ALIGN_RIGHT, "%i fps", game_FPS);
			al_draw_textf(default_text, al_map_rgb(100, 100, 100), display_width, fps_text_size + 10, ALLEGRO_ALIGN_RIGHT, "ex time: %.1f", al_current_time());

			al_draw_textf(default_text, al_map_rgb(100, 100, 100), 0, display_height - default_text_size - 10, ALLEGRO_ALIGN_LEFT, "%.0f %% of sample speed", dinamic_pitch*100);

            al_draw_bitmap(cursor, pos_x - cursor_rad, pos_y - cursor_rad, 0);

			al_flip_display();
			al_clear_to_color(background_color);
		}
	}

	cout << endl << "Free up memory...\n";

	// FREE UP MEMORY
	al_destroy_config(config_file);

	al_destroy_display(display);

	al_destroy_event_queue(event_queue);
	al_destroy_timer(fps_timer);

	al_destroy_font(default_text);
	al_destroy_font(fps_text);

	al_destroy_bitmap(original_cursor);
	al_destroy_bitmap(cursor);

	al_destroy_sample(song);
	al_destroy_sample_instance(song_instance);
	for (int i = 0; i < 8; i++)
	{
		al_destroy_sample(sample[i]);
		al_destroy_sample_instance(sample_instance[i]);
	}

	cout << "Exit\n";

	return 0;
}
